# coding=utf-8
#
# config.py
# Fira Core - Config
#
# Created by Marquis Kurt on 04/01/20.
# Copyright © 2020 Marquis Kurt. All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
"""This submodule contains the configuration system for reading worlds."""

import os
import toml
from .data import CSWorldDataGenerator

class CSWorldConfigGenerateError(Exception):
    """Could not generate the world from the requested file or configuration."""

class CSWorldConfigReader(object):
    """The world configuration reader.

    The world configuration reader parses a TOML file and creates an object that stores information
        about the world that can be used to generate a map.

    Attributes:
        title: The title of the map.
        checks: A list of strings containing the requirements for completing the puzzle.
        allowed: A list containing the allowed blocks for a given world. Unnecessary if using
            Advanced Mode.
        data: A `CSWorldDataGenerator` containing the world fata from the generated map that
            can be used to generate a world.
    """

    _world_str = """"""

    title = ""
    checks = []
    allowed = []
    data = CSWorldDataGenerator("")

    def __init__(self, filepath="", **kwargs):
        # type: (CSWorldConfigReader, str, dict) -> None
        """Construct the configuration reader.

        Arguments:
            filepath: The path to the configuration file to generate the world from. Defaults to
                an empty string.
            **kwargs: Arbitrary keyword arguments.

        Kwargs:
            title (str): The title of the map.
            checks (list): A list containing the checks for this particular level.
            allowed (list): A list containing the allowed blocks in basic mode.
            world (str): A string representation of the world layout.
            exists (callable): The function to use, if not relying on the built-in `os` module
                to determine whether the configuration file path is loadable.
            load (callable): The function to use, if not relying on the the built-in `open`
                function to load the file object.
        """
        config = {}

        if not filepath and not kwargs:
            raise CSWorldConfigGenerateError("Cannot generate an empty configuration.")

        if filepath:

            exists_fn = os.path.isfile
            load_fn = lambda a: open(a, 'r')

            if "exists" in kwargs and callable(kwargs["exists"]):
                exists_fn = kwargs["exists"]

            if not exists_fn(filepath):
                raise IOError("Cannot locate file %s" % (filepath))

            if "load" in kwargs and callable(kwargs["load"]):
                load_fn = kwargs["load"]

            with load_fn(filepath) as file_object:
                config = toml.load(file_object)

            if "level" not in config:
                raise CSWorldConfigGenerateError("Missing key 'level' in config.")

            current_level = config["level"]

            if "map" not in current_level and "config" not in current_level:
                raise CSWorldConfigGenerateError("Missing configuration details.")

            lvl_config = current_level["config"]
            lvl_map = current_level["map"]

            self.title = lvl_config["name"]
            self.checks = lvl_config["check"]
            self.allowed = lvl_config["allowed_blocks"]
            self._world_str = lvl_map["layout"]

        if "title" in kwargs:
            self.title = kwargs["title"]

        if "checks" in kwargs:
            self.checks = kwargs["checks"]

        if "allowed" in kwargs:
            self.allowed = kwargs["allowed"]

        if "world" in kwargs:
            self._world_str = kwargs["world"]

        self.data = CSWorldDataGenerator(self._world_str)



# coding=utf-8
#
# data.py
# Fira Core - Data
#
# Created by Marquis Kurt on 04/02/20.
# Copyright © 2020 Marquis Kurt. All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
"""This submodule contains the world data generator code for the configuration system."""

from .grid import CSGrid

class CSWorldDataGenerateError(Exception):
    """Could not generate the world data."""

class CSWorldDataGenerator(object):
    """The base class for the world data generator for world maps.

    The world data generator reads the given input string and returns a grid-like two-dimensional
        array for use with world manipulation or storage.
    """

    _str = ""
    _data = []
    _dimensions = 0, 0

    def __init__(self, data=""):
        # type: (CSWorldDataGenerator, str) -> None
        """Construct the world data generator.

        Arguments:
            data: The string containing the world's data to parse. Defaults to an empty string.
        """
        self._str = data

        total_rows = total_columns = 0
        current_column = 0
        current_row = []
        world_matrix = []
        did_set_player = False
        did_set_exit = False
        player_position = 0, 0

        world_map_terrain = {
            "%": "WALL",
            "P": "PLAYER",
            "E": "EXIT",
            ".": "COIN"
        }

        for line in data:
            if line == "\n":
                world_matrix.append(current_row[:])
                current_column = 0
                total_rows += 1
                current_row = []
            else:
                current_tile = world_map_terrain.get(line, "AIR")
                current_row.append(current_tile)

                if current_tile == "EXIT":
                    if did_set_exit:
                        raise CSWorldDataGenerateError("Cannot instantiate more than one exit.")

                if current_tile == "PLAYER":
                    if did_set_player:
                        raise CSWorldDataGenerateError("Cannot instantiate more than one player \
                                                    at coordinates %s and %s."
                                                       % (player_position,
                                                          (total_rows, current_column)))

                    self._player_position = len(world_matrix), current_column - 1

                if total_rows == 0:
                    total_columns += 1
                current_column += 1

        self._data = world_matrix
        self._dimensions = total_rows, total_columns

    def __str__(self):
        return ("Dimensions: %s\nMap\n=====\n" + self._str) % (self._dimensions)

    def size(self):
        # type: (CSWorldDataGenerator) -> tuple[int, int]
        """Get the size of the given world data grid.

        Returns:
            A tuple containing the number of rows and columns in the grid.
        """
        return self._dimensions

    def to_grid(self):
        # type: (CSWorldDataGenerator) -> CSGrid
        """Get the world data as a world grid.

        Returns:
            A world grid containing the world data.
        """
        return CSGrid(self._data, grid_filter=None)

    def coins(self):
        # type: (CSWorldDataGenerator) -> CSGrid
        """Get the world coin data.

        Returns:
            grid (CSGrid): The world grid containing the coins.
        """
        return CSGrid(self._data, grid_filter=lambda x: x == "COIN")

    def walls(self):
        # type: (CSWorldDataGenerator) -> CSGrid
        """Get the world wall data.

        Returns:
            grid (CSGrid): The world grid containing the walls.
        """
        return CSGrid(self._data, grid_filter=lambda x: x == "WALL")



# coding=utf-8
#
# grid.py
# Fira API - Grid
#
# Created by Marquis Kurt on 04/04/20.
# Copyright © 2020 Marquis Kurt. All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
"""The grid submodule contains a public-facing version of the internal grid used in the Unscripted
    minigame core.

The world grid is a two-dimensional array (list of lists) that contain strings that determine what
    element is present at the specific row and column. The main class responsible for working with
    the grid, `CSWorldGrid`, is a clean and lightweight implementation of this grid with multiple
    utilities to manage it. The grid implementation is not dependent on the `numpy` library and is
    geared towards accessing elements and other information, rather than mathematical operations.
"""
from typing import Callable, Optional

from ..core import CSGrid

# To expose the documentation for the grid without exposing the core, a subclassed version is
# created here. pdoc will automatically use the documentation based on class inheritance rules.

class CSWorldGrid(CSGrid):      #pylint:disable=missing-class-docstring

    def __init__(self, grid, filter=None):  #pylint:disable=redefined-builtin
        # type: (CSWorldGrid, list, Optional[Callable[[any], any]]) -> None
        CSGrid.__init__(self, grid, filter)

    def __str__(self):
        return CSGrid.__str__(self)

    def __eq__(self, value):
        return CSGrid.__eq__(self, value)

    def __ne__(self, value):
        return CSGrid.__ne__(self, value)

    def shape(self):
        # type: (CSWorldGrid) -> tuple[int, int]
        return CSGrid.shape(self)

    def as_list(self):
        # type: (CSWorldGrid) -> list[any]
        return CSGrid.as_list(self)

    def first(self, of=""):
        # type: (CSWorldGrid, str) -> any
        return CSGrid.first(self, of)

    def last(self, of=""):
        # type: (CSWorldGrid, str) -> any
        return CSGrid.last(self, of)

    def element_at(self, row, column):
        # type: (CSWorldGrid, int, int) -> any
        return CSGrid.element_at(self, row, column)



# coding=utf-8
#
# vm.py
# Fira Core - VM
#
# Created by Marquis Kurt on 04/03/20.
# Copyright © 2020 Marquis Kurt. All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
"""This submodule contains the low-level virtual machine reader for the minigame
code.

## Implementation

The Unscripted API and the minigame GUI eventually write files readable by the
Nadia Virtual Machine (NVM). The Nadia Virtual Machine is a stack-based virtual
machine designed to perform operations specific to the Unscripted minigame.

### Commands
In the NadiaVM, there are several commands:

- `alloc (str) (int)`: Create a new array with a name and a specific size.
- `collect`: Collect a coin in the world. Acts like a pause in VM.
- `push (str) (int)`: Push the top of the stack to the given array at the index.
- `pop (str) (int)`: Pop the item located at the array's index and move it to the
    top of the stack.
- `set constant (any)`: Set the top of the stack to a constant value.
- `move player (str)`: Move the player in a given direction.
- `exit player`: Try to end the execution script and finish the level.
"""
from typing import Optional


class CSNadiaVMCommandNotFoundError(Exception):
    """VM command not found."""

class CSNadiaVM(object):
    """An implementation of the NadiaVM stack."""

    _instructions = []
    _stack = []

    def __init__(self, path, pl):
        # type: (CSNadiaVM, str, tuple[int, int]) -> None
        """Construct the VM reader.

        Arguments:
            path (str): The path to the compiled NadiaVM file (`.nvm`) to read from.
            pl (tuple): The coordinates of the player.
        """
        with open(path, 'r') as vm_file:
            self._instructions = list(filter(lambda a: a, vm_file.read().split("\n")))
        self._player_pos = pl

    def has_more_instructions(self):
        # type: (CSNadiaVM) -> bool
        """Determine if the VM has more instructions to execute.

        Returns:
            more (bool): Boolean that will be True if there are more instructions, False otherwise.
        """
        return len(self._instructions) > 0

    def preview_next_instruction(self):
        # type: (CSNadiaVM) -> Optional[str]
        """Get the next command in the instruction list without executing it in the VM.

        Returns:
            command (str): The command to be executed, excluding parameters, or None if there
                are no more instructions to execute in the VM.
        """
        if self.has_more_instructions():
            return self._instructions[0].split(" ")[0]
        else:
            return None

    def next(self):
        # type: (CSNadiaVM) -> None
        """Execute the next instruction in the VM code.

        Raises:
            error (CSNadiaVMCommandNotFoundError): Command not found.
        """
        current = self._instructions.pop(0).split(" ")

        if current[0] == "alloc":
            self._alloc(current[1], int(current[2]))
        elif current[0] == "push":
            self._push(current[1], int(current[2]))
        elif current[0] == "pop":
            self._pop(current[1], int(current[2]))
        elif current[0] == "set":
            self._set(current[2])
        elif current[0] == "move":
            self._move(current[2])
        elif current[0] == "collect":
            pass
        elif current[0] != "exit":
            raise CSNadiaVMCommandNotFoundError("Invalid command: '%s'" % (current[0]))

    def _alloc(self, name, size):
        """Allocate a space of memory for a given array.

        Arguments:
            name (str): The name of the array to allocate space for.
            size (int): The size of the array. Defaults to 1.
        """
        self.__dict__[name] = [None for x in range(size)]

    def _set(self, value):
        """Set the top of the stack to a constant value.

        Arguments:
            value (any): The value to create a constant for.
        """
        self._stack.append(value)

    def _push(self, name, index):
        """Push the top-most item on the current stack to the given array.

        Arguments:
            name (str): The name of the array to push to.
            index (int): The index of the array to push to.
        """
        self.__dict__[name][index] = self._stack.pop()

    def _pop(self, name, index):
        """Pop the item from the array at a given index and set it at the top
        of the execution stack.

        Arguments:
            array (str): The array to pop an item from.
            index (int): The index of the item in the array to pop.
        """
        self._stack.append(self.__dict__[name][index])
        self.__dict__[name][index] = None

    def _move(self, direction):
        """Move the player in a given direction.

        Arguments:
            direction (str): The direction the player will move in.
        """
        transforms = {
            "north": (-1, 0),
            "south": (1, 0),
            "west": (0, -1),
            "east": (0, 1)
        }

        trans_x, trans_y = transforms.get(direction, "east")
        curr_x, curr_y = self._player_pos
        self._player_pos = curr_x + trans_x, curr_y + trans_y

    def get(self, name):
        # type: (CSNadiaVM, str) -> Optional[list]
        """Get the specified item in the virtual machine.

        Arguments:
            name (str): The name of the item to get.

        Returns:
            array (list): The specified item, or None if it doesn't exist.
        """
        return self.__dict__[name] if name in self.__dict__ \
            and type(self.__dict__[name]) is list else None #pylint:disable=unidiomatic-typecheck

    def pos(self):
        # type: (CSNadiaVM) -> tuple[int, int]
        """Get the current player position from the VM execution stack.

        Returns:
            player (tuple): A tuple containing the coordinates of the player.
        """
        return self._player_pos

class CSNadiaVMWriterBuilder(object):
    """An list-based implementation of the NadiaVM file writer.

    This class is similar to CSNadaVMWriter and contains the same methods; however,
        CSNadiaVMWriterBuilder uses a list to store its code rather the string that CSNadiaVMWriter
        uses. This is useful in instances where the builder needs to remove pieces of code or work
        with the current set of instructions as a list.

    Attributes:
        instructions (list): The list of VM commands to write to the VM file.
    """

    instructions = []

    def __init__(self, path):
        # type: (CSNadiaVMWriterBuilder, str) -> None
        """Construct the VM writer builder.

        Arguments:
            path (str): The path to the compiled NadiaVM file (`.nvm`) to write to.
        """
        self.path = path

    def __str__(self):
        string = "Filepath: %s\n" % (self.path)
        for instruction in self.instructions:
            string += "%s\n" % (instruction)
        return string

    def alloc(self, array_name, size=1):
        # type: (CSNadiaVMWriterBuilder, str, int) -> None
        """Allocate a space of memory for a given array.

        Arguments:
            array_name (str): The name of the array to allocate space for.
            size (int): The size of the array. Defaults to 1.
        """
        self.instructions.append("alloc %s %s\n" % (array_name, size))

    def push(self, array, index):
        # type: (CSNadiaVMWriterBuilder, str, int) -> None
        """Push the top-most item on the current stack to the given array.

        Arguments:
            array (str): The name of the array to push to.
            index (int): The index of the array to push to.
        """
        self.instructions.append("push %s %s\n" % (array, index))

    def pop(self, array, index):
        # type: (CSNadiaVMWriterBuilder, str, int) -> None
        """Pop the item from the array at a given index and set it at the top
        of the execution stack.

        Arguments:
            array (str): The array to pop an item from.
            index (int): The index of the item in the array to pop.
        """
        self.instructions.append("pop %s %s\n" % (array, index))

    def set(self, value):
        # type: (CSNadiaVMWriterBuilder, any) -> None
        """Set the top of the stack to a constant value.

        Arguments:
            value (any): The value to create a constant for.
        """
        self.instructions.append("set constant " + str(value) + "\n")

    def move(self, direction):
        # type: (CSNadiaVMWriterBuilder, str) -> None
        """Move the player in a given direction.

        Arguments:
            direction (str): The direction the player will move in.
        """
        self.instructions.append("move player " + direction + "\n")

    def collect(self):
        # type: (CSNadiaVMWriterBuilder) -> None
        """Collect. In the VM, this acts like a pause."""
        self.instructions.append("collect\n")

    def exit(self):
        # type: (CSNadiaVMWriterBuilder) -> None
        """Try to exit the world and end execution of the script."""
        self.instructions.append("exit player\n")

    def write(self):
        # type: (CSNadiaVMWriterBuilder) -> None
        """Write the VM code to the requested file."""
        with open(self.path, 'w+') as vm_file_stream:
            vm_file_stream.write("".join(self.instructions))

    def clear(self):
        # type: (CSNadiaVMWriterBuilder) -> None
        """Clear all of the current instructions in the VM stack."""
        del self.instructions[:]

    def undo(self, ignore_collect=True):
        # type: (CSNadiaVMWriterBuilder, bool) -> None
        """Remove the top of the instruction stack.

        Arguments:
            ignore_collect (bool): Whether to ignore the pop and push statements preceding the
                collect statement. Defaults to True.
        """
        if len(self.instructions) > 0:
            instruction = self.instructions.pop()

            if instruction == "collect" and not ignore_collect:
                self.instructions.pop()     # Pops the push statement.
                self.instructions.pop()     # Pops the pop statement.

class CSNadiaVMWriter(object):
    """An implementation of the NadiaVM file writer."""

    code = """"""

    def __init__(self, path):
        # type: (CSNadiaVMWriter, str) -> None
        """Construct the VM writer.

        Arguments:
            path (str): The path to the compiled NadiaVM file (`.nvm`) to write to.
        """
        self.path = path

    def __str__(self):
        return "Filepath: " + self.path + "\n" + self.code

    def alloc(self, array_name, size=1):
        # type: (CSNadiaVMWriter, str, int) -> None
        """Allocate a space of memory for a given array.

        Arguments:
            array_name (str): The name of the array to allocate space for.
            size (int): The size of the array. Defaults to 1.
        """
        self.code += "alloc %s %s\n" % (array_name, size)

    def push(self, array, index):
        # type: (CSNadiaVMWriter, str, int) -> None
        """Push the top-most item on the current stack to the given array.

        Arguments:
            array (str): The name of the array to push to.
            index (int): The index of the array to push to.
        """
        self.code += "push %s %s\n" % (array, index)

    def pop(self, array, index):
        # type: (CSNadiaVMWriter, str, int) -> None
        """Pop the item from the array at a given index and set it at the top
        of the execution stack.

        Arguments:
            array (str): The array to pop an item from.
            index (int): The index of the item in the array to pop.
        """
        self.code += "pop %s %s\n" % (array, index)

    def set(self, value):
        # type: (CSNadiaVMWriter, any) -> None
        """Set the top of the stack to a constant value.

        Arguments:
            value (any): The value to create a constant for.
        """
        self.code += "set constant " + str(value) + "\n"

    def move(self, direction):
        # type: (CSNadiaVMWriter, str) -> None
        """Move the player in a given direction.

        Arguments:
            direction (str): The direction the player will move in.
        """
        self.code += "move player " + direction + "\n"

    def collect(self):
        # type: (CSNadiaVMWriter) -> None
        """Collect. In the VM, this can act like a pause."""
        self.code += "collect\n"

    def exit(self):
        # type: (CSNadiaVMWriter) -> None
        """Try to exit the world and end execution of the script."""
        self.code += "exit player\n"

    def write(self):
        # type: (CSNadiaVMWriter) -> None
        """Write the VM code to the requested file."""
        with open(self.path, 'w+') as vm_file_stream:
            vm_file_stream.write(self.code)



# coding=utf-8
#
# info.py
# Fira API - Info
#
# Created by Marquis Kurt on 04/04/20.
# Copyright © 2020 Marquis Kurt. All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
"""The info submodule contains the utilities to get the world and player information for a given
    level."""

from .player import CSPlayer
from .world import CSWorld
from ..core import CSWorldConfigReader
from ..core import CSNadiaVMWriter

def get_level_information(level, fn_path="", **kwargs):
    # type: (int, str, dict) -> tuple[CSPlayer, CSWorld]
    """Create a world and player based on a game level.

    Arguments:
        level (int): The level number as indicated by the minigame.
        fn_path (str): The path to where the NadiaVM file will be written to. This excludes
            the file name itself.
        **kwargs (dict): Arbitrary keyword arguments.

    Kwargs:
        config_file (str): The path to the level configuration file, excluding the file name.
        exists (callable): The function to use, if not relying on the built-in `os` module
            to determine whether the configuration file path is loadable.
        load (callable): The function to use, if not relying on the the built-in `open`
            function to load the file object.

    Returns:
        info (tuple): A tuple containing the `minigame.api.player.CSPlayer` object and the
            `minigame.api.world.CSWorld` object
    """
    conf = kwargs["config_file"] + "/level%s.toml" % (toml) \
        if "config_file" in kwargs \
            else "core/minigame/levels/level%s.toml" % (level)
    w_info = CSWorldConfigReader(conf, **kwargs)
    writer = CSNadiaVMWriter(fn_path + "/compiled/lvl%s.nvm" % (level))
    world = CSWorld(from_data=w_info.data)
    player = CSPlayer(in_world=world, vm=writer)
    return player, world



# coding=utf-8
#
# __init__.py
# Fira API
#
# Created by Marquis Kurt on 04/01/20.
# Copyright © 2020 Marquis Kurt. All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
"""The `api` module contains all of the player-facing functions and classes used in the
    Advanced Mode of the Unscripted minigame.

The Unscriped Minigame Application Programming Interface (API) allows players to directly
    manipulate the player's movements and actions while traversing in the minigame world using
    Python. The minigame's logic will handle translating any commands to display at the end of the
    script compilation.

.. WARNING::
   Although most (if not all) of the submodules in the parent package are publicly accessible,
    it is heavily encouraged that players make use of the standard utilities provided with
    the API module.

   The documentation for some components in the core submodule are provided for reference namely,
    `minigame.api.grid.CSWorldGrid`.

The `api` module comes with a few submodules that contain function and utilities for multiple
    aspects of the minigame:

- `player` hosts all of the code related to manipulating the player in the minigame.
- `world` hosts all of the code related to viewing and gathering information about the world
    in the minigame.
- `grid` hosts a publicly available version of the internal grid system used for world generation.
"""
from .world import CSWorld
from .grid import CSWorldGrid
from .player import CSPlayer
from .info import get_level_information

__version__ = "1.2.0"



# coding=utf-8
#
# player.py
# Fira API - Player
#
# Created by Marquis Kurt on 03/31/20.
# Copyright © 2020 Marquis Kurt. All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
"""The player module contains the functions and classes necessary to manipulate players in
    a minigame world.

The module contains a class that lets players control the behavior of their minigame counterpart
    by moving in a given direction, collecting an item, etc. (`CSPlayer`). To prevent accidental
    manipulation of the original world data, the player class uses its own position property to
    update its location.
"""
from .world import CSWorld
from ..core import CSNadiaVMWriter

class CSPlayer(object):
    """The base class for a player in the minigame world.

    The player object contains methods for manipulating the current player's
        position and inventory system.
    """
    _position = (0, 0)
    _inventory = []
    _world = None
    _world_coins = []
    _vm = None

    def __init__(self, in_world, **kwargs):
        # type: (CSPlayer, CSWorld, dict) -> None
        """Construct the Player object.

        Arguments:
            in_world (minigame.api.world.CSWorld): The world the player is located in.
            **kwargs (dict): Arbitrary keyword arguments.

        Kwargs:
            at_position (tuple): The position the player should be placed in. Defaults to the player
                position in the world (`minigame.api.world.CSWorld.player`).
            with_inv (list): A list containing items that the player will have to start. Defaults to
                an empty list.
            vm: The virtual machine writer, if available.
        """
        if not isinstance(in_world, CSWorld):
            raise TypeError("Expected a minigame world type but received %s instead."
                            % (type(in_world)))
        self._world = in_world
        self._world_coins = self._world.coins().as_list()[:]

        self._position = kwargs['at_position'] if "at_position" in kwargs else self._world.player()
        self._inventory = kwargs['with_inv'] if "with_inv" in kwargs else []

        if "vm" in kwargs:
            self._vm = kwargs["vm"]
            self._vm.alloc("world_coins", len(self._world_coins))
            self._vm.alloc("inventory", len(self._world_coins))

            for coin in self._world_coins:
                self._vm.set(coin)
                self._vm.push("world_coins", self._world_coins.index(coin))

    def location(self):
        # type: (CSPlayer) -> tuple[int, int]
        """Get the player's current position.

        Returns:
            position (tuple): The current coordinates of the player.
        """
        return self._position

    def origin(self):
        # type: (CSPlayer) -> tuple[int, int]
        """Get the original starting position of the player.

        Returns:
            origin (tuple): The coordinates of the player's original position.
        """
        return self._world.player()

    def capacity(self):
        # type: (CSPlayer) -> int
        """Get the the count of how many items the player has.

        Returns:
            count (int): The number of items in the inventory.
        """
        return len(self._inventory)

    def blocked(self):
        # type: (CSPlayer) -> bool
        """Determine whether a player is blocked at a given position.

        Returns:
            blocked (bool): True if any walls are near the player (1-block radius).
        """
        px, py = self._position #pylint:disable=invalid-name

        positions = [(px-1, py), (px+1, py), (px, py-1), (px, py+1)]

        for position in positions:
            if position in self._world.walls().as_list():
                return True
        return False

    def move(self, direction):
        # type: (CSPlayer, str) -> CSPlayer
        """Move the player in a direction, if the direction results in the player
            being able to move into a non-walled area.

        Arguments:
            direction (str): The direction the player should move in. Acceptable directions
                are `"north"`, `"south"`, `"east"`, and `"west"`.

        Returns:
            player (CSPlayer): The Player that committed the move action. This is useful in cases
                where chaining methods is preferred.
        """
        transforms = {
            "north": (-1, 0),
            "south": (1, 0),
            "west": (0, -1),
            "east": (0, 1)
        }

        trans_x, trans_y = transforms.get(direction, "east")
        curr_x, curr_y = self._position
        new_x, new_y = curr_x + trans_x, curr_y + trans_y

        if (new_x, new_y) not in self._world.walls().as_list():
            self._position = new_x, new_y

        if self._vm:
            self._vm.move(direction)

        return self

    def collect(self):
        # type: (CSPlayer) -> CSPlayer
        """Add an item into the player's inventory at the player's current position.

        If the item does not exist in the world, or the player already has the item in question,
            nothing occurs.

        Returns:
            player (CSPlayer): The Player object that committed the collect action. This is useful
                in cases where chaining methods is preferred.
        """
        item_index = self._world_coins.index(self._position)
        self._inventory.append(self._position)

        if self._vm:
            self._vm.pop("world_coins", item_index)
            self._vm.push("inventory", item_index)
            self._vm.collect()

        self._world_coins[item_index] = None
        return self

    def exit(self):
        # type: (CSPlayer) -> None
        """Exit the level, if possible.

        If a VM is specified, the VM writer will also close the writer by writing to the VM file.
        """
        if self._vm:
            self._vm.exit()
            self._vm.write()



# coding=utf-8
#
# template.py
# Fira Core - Template
#
# Created by Marquis Kurt on 04/06/20.
# Copyright © 2020 Marquis Kurt. All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
"""The template module contains the utilities to create file templates for the Unscripted
    minigame code.

The template files are generated when the game starts and are replaced if none are found.
"""

def generate_template(filepath, for_level=0):
    # type: (str, int) -> None
    """Generate a template file using the Minigame APIs.

    Arguments:
        filepath (str): The path to where the template file will be written.
        for_level (int): The corresponding level for the minigame. Defaults to 0.
    """
    template = """#
# This script corresponds to the Advanced Mode script for Level %s in the minigame.
# It is recommended to keep the appropriate template code to start. Remember that the goal is to
# collect all coins and reach the exit.
#
# To access the documentation for the minigame APIs, either go to Settings > Minigame and click
# "Open Documentation", go to Help > Documentation, or visit the following link in your browser:
# https://fira.marquiskurt.net/api/. Go to https://fira.marquiskurt.net/gameplay.html#limitations-1
# to view the limitations of using the official API as called here.
#
# If you want to use a third-party tool or framework instead of the official Fira API and want to
# sideload in a virtual machine file, do not write any code in this file. More information can be
# found at https://fira.marquiskurt.net/implementation.html#obtaining-code-from-third-party-tools.
#

# Import the level information APIs.
from uvn_fira.api import get_level_information, CSPlayer, CSWorld

# Get all of the information for this particular level.
game_player, game_world = get_level_information(%s,
                                                fn_path=renpy.config.savedir + "/minigame/",
                                                exists=renpy.loadable,
                                                load=renpy.exports.file) # type: (CSPlayer, CSWorld)

# WRITE CODE HERE
""" % (for_level, for_level)
    with open(filepath, 'w+') as file_obj:
        file_obj.write(template)



# coding=utf-8
#
# __init__.py
# Fira
#
# Created by Marquis Kurt on 03/31/20.
# Copyright © 2020 Marquis Kurt. All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#

"""The `uvn_fira` package contains all of the source code for the Unscripted minigame."""
from uvn_fira.api import *
from uvn_fira.core import *

__version__ = "1.2.0"



# coding=utf-8
#
# grid.py
# Fira Core - Grid
#
# Created by Marquis Kurt on 04/04/20.
# Copyright © 2020 Marquis Kurt. All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
"""This submodule contains the grid data type used in the minigame world."""
from typing import Optional, Callable

class CSGrid(object):
    """A class representation of a grid.

    The grid is a two-dimensional array object that contains elements at given positions. This
        class contains methods that allow easy access to the contents of the grid without needing
        to access the grid directly.

    Attributes:
        grid (list): The two-dimensional array containing items at given positions,
            organized by row and column.

    """
    grid = []

    def __init__(self, grid, grid_filter=None):
        # type: (CSGrid, list, Optional[Callable[[any], any]]) -> None
        """Construct a grid.

        If a filter is supplied with the grid, the grid will filter for the items specified in the
            filter expression, marking those that don't match the filter as None.

        Arguments:
            grid (list): The two-dimensional array representing the grid.
            grid_filter (callable): A function that determines what items in the grid to keep.
                Defaults to None, indicating that no filter is applied.
        """
        temp_grid = []
        if callable(grid_filter):
            for row in grid:
                temp_grid.append([x if grid_filter(x) else None for x in row])
        else:
            temp_grid = grid[:]
        self.grid = temp_grid

    def __str__(self):
        data = ""
        for row in self.grid:
            data += str(row).replace("[", "").replace("]", "").replace(",", "\t") + "\n"
        return data

    def __eq__(self, value):
        return isinstance(value, CSGrid) and self.grid == value.grid

    def __ne__(self, value):
        return not self.__eq__(value)

    def shape(self):
        # type: (CSGrid) -> tuple[int, int]
        """Get the shape of the grid.

        Returns:
            shape (tuple): A tuple containig the total rows and columns of this grid.
        """
        rows = len(self.grid)
        columns = len(self.grid[0])
        return rows, columns

    def as_list(self):
        # type: (CSGrid) -> list[any]
        """Convert the grid to a list of coordinates containing a non-void item.

        .. NOTE::
           If a filter was applied to the grid at construction, it will only select items
           that are not `None`.

        Returns:
            coordinates (list): A list of tuples containing the coordinates to valid items in the
                grid.
        """
        coordinates = []
        for row in range(len(self.grid) - 1):
            for column in range(len(self.grid[0]) - 1):
                if self.grid[row][column] is not None:
                    coordinates.append((row, column))
        return coordinates

    def first(self, of=""):     #pylint:disable=invalid-name
        # type: (CSGrid, str) -> tuple[int, int]
        """Get the first instance of an item in the grid.

        Arguments:
            of (str): The item to look for the first instance of in this grid.

        Returns:
            coords (tuple): A tuple containing the row and column coordinates of the first item. If
                the item was not found, the tuple `-1, -1` is returned.
        """
        coords = (-1, -1)

        for row in range(len(self.grid) - 1):
            for column in range(len(self.grid[0]) - 1):
                if self.grid[row][column] == of:
                    coords = row, column
                    break
        return coords

    def last(self, of=""):      #pylint:disable=invalid-name
        # type: (CSGrid, str) -> tuple[int, int]
        """Get the last instance of an item in the grid.

        Arguments:
            of (str): The item to look for the last instance of in this grid.

        Returns:
            coords (tuple): A tuple containing the row and column coordinates of the last item. If
                the item was not found, the tuple `-1, -1` is returned.
        """
        inverse = [row[::-1] for row in self.grid][::-1]
        coords = -1, -1

        for row in range(len(inverse) - 1):
            for column in range(len(self.grid[0]) - 1):
                if inverse[row][column] == of:
                    coords = row, column
                    break

        return coords

    def element_at(self, row, column):
        # type: (CSGrid, int, int) -> Any
        """Get the element at a specified position.

        Arguments:
            row (int): The row that the element is located in
            column (int): The column that the element is located in

        Returns:
            element (str): The element at the specified position in the grid.
        """
        return self.grid[row][column]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['uvn_fira', 'uvn_fira.api', 'uvn_fira.core']

package_data = \
{'': ['*']}

install_requires = \
['toml>=0.10.0,<0.11.0', 'typing>=3.7.4,<4.0.0']

setup_kwargs = {
    'name': 'uvn-fira',
    'version': '1.2.10',
    'description': 'The backend and API code for the Unscripted mini-game.',
    'long_description': '# Fira\n\n**Fira** is the main backend and API code for the minigame in [Unscripted](https://unscripted.marquiskurt.net), a visual novel about software development. Fira provides many facets of the minigame, including a public API that players can use to code solutions to the minigame puzzles, a configuration and data generator from level files, and a virtual machine that runs low-level code that the minigame processes. Fira is named after Fira Sans, one of the game\'s characters.\n\n\n[![MPL](https://img.shields.io/github/license/alicerunsonfedora/fira)](LICENSE.txt)\n![Python](https://img.shields.io/badge/python-2.7+-blue.svg)\n[![PyPI version](https://badge.fury.io/py/uvn-fira.svg)](https://pypi.org/project/uvn-fira)\n\n## Requirements\n\n- Python 2.7+\n- Poetry package manager\n\n## Getting started\n\nFira comes pre-packaged in Unscripted but can be installed outside of the game to work better with IDEs and other Python tools such as Poetry.\n\n### Dependencies\n\nFira is both a Python 2 and Python 3 package and relies on the TOML Python package. These dependencies will be installed with the package, either from source or from PyPI.\n\n### Quick Start: Install on PyPI\n\nFira is available on PyPI and can be installed as such:\n\n```\npip install uvn-fira\n```\n\n### Install from source\n\nTo install Fira from the source code, first clone the repository from GitHub via `git clone`. You\'ll also need to install [Poetry](https://python-poetry.org). In the root of the source, run the following commands:\n\n```\npoetry install\npoetry build\n```\n\nThe resulting wheel files will be available in the `dist` directory.\n\n## Usage\n\nFor players installing this package to solve minigame puzzles, using the Fira package to access the API is relatively straightforward:\n\n```py\nfrom uvn_fira.api import get_level_information, CSPlayer, CSWorld\n\ngp, gw = get_level_information(0,\n                               fn_path=renpy.config.savedir + "/minigame",\n                               exists=renpy.loadable,\n                               load=renpy.exports.file) # type: (CSPlayer, CSWorld)\n```\n\nDocumentation on the API is located inside of Unscripted by going to **Help &rsaquo; Minigame** or **Settings &rsaquo; Minigame**.\n\nThe documentation for the entire package is located at [https://fira.marquiskurt.net](https://fira.marquiskurt.net), which is useful for developers that wish to make custom toolkits that connect to the minigame\'s virtual machine or for modders that wish to make custom minigame levels.\n\n## Reporting bugs\n\nBugs and feature requests for Fira can be submitted on YouTrack. [Submit a report &rsaquo;](https://youtrack.marquiskurt.net/youtrack/newIssue?project=FIRA)\n\n## License\n\nThe Fira package is licensed under the Mozilla Public License v2.0.',
    'author': 'Marquis Kurt',
    'author_email': 'software@marquiskurt.net',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/UnscriptedVN/fira',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, !=3.6.*',
}


setup(**setup_kwargs)



# coding=utf-8
#
# __init__.py
# Fira Core
#
# Created by Marquis Kurt on 04/02/20.
# Copyright © 2020 Marquis Kurt. All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
"""The `core` module contains the utilities, functions, and classes that control the primary logic
    for the Unscripted minigame.
"""

from .config import *
from .data import *
from .grid import *
from .template import *
from .vm import *

__version__ = "1.2.0"



# coding=utf-8
#
# world.py
# Fira API - World
#
# Created by Marquis Kurt on 03/31/20.
# Copyright © 2020 Marquis Kurt. All rights reserved.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
#
"""The world module contains all of the classes and functions necessary to visualize
    and gather information about the game world.

The module contains the base grid class responsible for making two-dimensional arrays
    organized like a grid easier to work with (`minigame.api.grid.CSWorldGrid`), as well as the
    world class that contains all of the world's information such as the player's position and
    coins (`CSWorld`).

## World Implementation
Worlds are organized as a grid-like, two-dimensional array (matrix) consisting of rows and columns.
    Each element in the grid contains a string that describes what the world element at that
    position in the grid is. There are five possible options:

- `"PLAYER"`, which refers to the player character
- `"WALL"`, which refers to a wall
- `"COIN"`, which refers to a collectable coin
- `"EXIT"`, which refers to an exit
- `"AIR"`, which refers to a space that doesn't have a particular item

In most scenarios, the world itself does not get modified directly since the layout of items in the
    world need to be preserved. The `CSPlayer` object in the `player` module of the API handles
    world manipulation when a world is passed into its constructor, and most scripts will make a
    copy of world information such as the number of coins and the player's starting position.

It may be impractical in some cases to access an element in the grid directly. The
    `minigame.api.grid.CSWorldGrid` class allows for the world grid to be used in a more practical
    manner by making common operations such as getting a list containing coordinates of a specific
    type of item and getting the first instance of an item easier.
"""
from ..core import CSNadiaVMWriter, CSWorldDataGenerator
from .grid import CSWorldGrid

class CSWorld(object):
    """The base class for a minigame world.

    The minigame world contains a matrix containing the elements used to generate that world,
    as well as any specific world properties like coins and exit locations.
    """
    _dimensions = (0, 0)
    _player_position = (0, 0)
    _grid = CSWorldGrid([])
    _vm_author = CSNadiaVMWriter("null.nvm")

    def __init__(self, from_data, **kwargs):
        # type: (CSWorld, CSWorldDataGenerator, dict) -> None
        """Construct a World object.

        Arguments:
            from_data (CSWorldDataGenerator): The world generator used to create the world data.
            **kwargs: Arbitrary keyword arguments.
        """

        self._grid = from_data.to_grid()
        self._dimensions = from_data.size()
        self._player_position = from_data.to_grid().first("PLAYER")

        if "nvm" in kwargs:
            self._vm_author = CSNadiaVMWriter(kwargs["nvm"])

    def player(self):
        # type: (CSWorld) -> tuple[int, int]
        """Get the player's current location in the world.

        Returns:
            position (tuple): The current coordinates of the player.
        """
        return self._player_position

    def size(self):
        # type: (CSWorld) -> tuple[int, int]
        """Get the size of the world.

        Returns:
            dimensions (tuple): A tuple containing the number of rows and columns.
        """
        return self._dimensions

    def walls(self):
        # type: (CSWorld) -> CSWorldGrid
        """Get the grid of walls in the world.

        Returns:
            walls (minigame.api.grid.CSWorldGrid): Grid containing only the walls.
        """
        return CSWorldGrid(self._grid.grid, lambda a: a == "WALL")

    def coins(self):
        # type: (CSWorld) -> CSWorldGrid
        """Get the grid of coins in the world.

        Returns:
            coins (minigame.api.grid.CSWorldGrid): Grid containing only the coins.
        """
        return CSWorldGrid(self._grid.grid, lambda a: a == "COIN")

    def exit(self):
        # type: (CSWorld) -> any
        """Get the location of the exit.

        Returns:
            exit (tuple): A tuple containing the coordinates of the exit.
        """
        return self._grid.first("EXIT")
